# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides 2D and 3D scatter data item class.
"""

from __future__ import absolute_import

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "15/11/2017"

try:
    from collections import abc
except ImportError:  # Python2 support
    import collections as abc
import logging
import numpy

from ....utils.deprecation import deprecated
from ... import _glutils as glu
from ...plot._utils.delaunay import delaunay
from ..scene import function, primitives, utils

from ...plot.items import ScatterVisualizationMixIn
from .core import DataItem3D, Item3DChangedType, ItemChangedType
from .mixins import ColormapMixIn, SymbolMixIn
from ._pick import PickingResult


_logger = logging.getLogger(__name__)


class Scatter3D(DataItem3D, ColormapMixIn, SymbolMixIn):
    """Description of a 3D scatter plot.

    :param parent: The View widget this item belongs to.
    """

    # TODO supports different size for each point

    def __init__(self, parent=None):
        DataItem3D.__init__(self, parent=parent)
        ColormapMixIn.__init__(self)
        SymbolMixIn.__init__(self)

        noData = numpy.zeros((0, 1), dtype=numpy.float32)
        symbol, size = self._getSceneSymbol()
        self._scatter = primitives.Points(
            x=noData, y=noData, z=noData, value=noData, size=size)
        self._scatter.marker = symbol
        self._getScenePrimitive().children.append(self._scatter)

        # Connect scene primitive to mix-in class
        ColormapMixIn._setSceneColormap(self, self._scatter.colormap)

    def _updated(self, event=None):
        """Handle mix-in class updates"""
        if event in (ItemChangedType.SYMBOL, ItemChangedType.SYMBOL_SIZE):
            symbol, size = self._getSceneSymbol()
            self._scatter.marker = symbol
            self._scatter.setAttribute('size', size, copy=True)

        super(Scatter3D, self)._updated(event)

    def setData(self, x, y, z, value, copy=True):
        """Set the data of the scatter plot

        :param numpy.ndarray x: Array of X coordinates (single value not accepted)
        :param y: Points Y coordinate (array-like or single value)
        :param z: Points Z coordinate (array-like or single value)
        :param value: Points values (array-like or single value)
        :param bool copy:
            True (default) to copy the data,
            False to use provided data (do not modify!)
        """
        self._scatter.setAttribute('x', x, copy=copy)
        self._scatter.setAttribute('y', y, copy=copy)
        self._scatter.setAttribute('z', z, copy=copy)
        self._scatter.setAttribute('value', value, copy=copy)

        self._setColormappedData(self.getValueData(copy=False), copy=False)
        self._updated(ItemChangedType.DATA)

    def getData(self, copy=True):
        """Returns data as provided to :meth:`setData`.

        :param bool copy: True to get a copy,
                          False to return internal data (do not modify!)
        :return: (x, y, z, value)
        """
        return (self.getXData(copy),
                self.getYData(copy),
                self.getZData(copy),
                self.getValueData(copy))

    def getXData(self, copy=True):
        """Returns X data coordinates.

        :param bool copy: True to get a copy,
                          False to return internal array (do not modify!)
        :return: X coordinates
        :rtype: numpy.ndarray
        """
        return self._scatter.getAttribute('x', copy=copy).reshape(-1)

    def getYData(self, copy=True):
        """Returns Y data coordinates.

        :param bool copy: True to get a copy,
                          False to return internal array (do not modify!)
        :return: Y coordinates
        :rtype: numpy.ndarray
        """
        return self._scatter.getAttribute('y', copy=copy).reshape(-1)

    def getZData(self, copy=True):
        """Returns Z data coordinates.

        :param bool copy: True to get a copy,
                          False to return internal array (do not modify!)
        :return: Z coordinates
        :rtype: numpy.ndarray
        """
        return self._scatter.getAttribute('z', copy=copy).reshape(-1)

    def getValueData(self, copy=True):
        """Returns data values.

        :param bool copy: True to get a copy,
                          False to return internal array (do not modify!)
        :return: data values
        :rtype: numpy.ndarray
        """
        return self._scatter.getAttribute('value', copy=copy).reshape(-1)

    @deprecated(reason="Consistency with PlotWidget items",
                replacement="getValueData", since_version="0.10.0")
    def getValues(self, copy=True):
        return self.getValueData(copy)

    def _pickFull(self, context, threshold=0., sort='depth'):
        """Perform picking in this item at given widget position.

        :param PickContext context: Current picking context
        :param float threshold: Picking threshold in pixel.
            Perform picking in a square of size threshold x threshold.
        :param str sort: How returned indices are sorted:

            - 'index' (default): sort by the value of the indices
            - 'depth':  Sort by the depth of the points from the current
              camera point of view.
        :return: Object holding the results or None
        :rtype: Union[None,PickingResult]
        """
        assert sort in ('index', 'depth')

        rayNdc = context.getPickingSegment(frame='ndc')
        if rayNdc is None:  # No picking outside viewport
            return None

        # Project data to NDC
        xData = self.getXData(copy=False)
        if len(xData) == 0:  # No data in the scatter
            return None

        primitive = self._getScenePrimitive()

        dataPoints = numpy.transpose((xData,
                                      self.getYData(copy=False),
                                      self.getZData(copy=False),
                                      numpy.ones_like(xData)))

        pointsNdc = primitive.objectToNDCTransform.transformPoints(
            dataPoints, perspectiveDivide=True)

        # Perform picking
        distancesNdc = numpy.abs(pointsNdc[:, :2] - rayNdc[0, :2])
        # TODO issue with symbol size: using pixel instead of points
        threshold += self.getSymbolSize()
        thresholdNdc = 2. * threshold / numpy.array(primitive.viewport.size)
        picked = numpy.where(numpy.logical_and(
                numpy.all(distancesNdc < thresholdNdc, axis=1),
                numpy.logical_and(rayNdc[0, 2] <= pointsNdc[:, 2],
                                  pointsNdc[:, 2] <= rayNdc[1, 2])))[0]

        if sort == 'depth':
            # Sort picked points from front to back
            picked = picked[numpy.argsort(pointsNdc[picked, 2])]

        if picked.size > 0:
            return PickingResult(self,
                                 positions=dataPoints[picked, :3],
                                 indices=picked,
                                 fetchdata=self.getValueData)
        else:
            return None


class Scatter2D(DataItem3D, ColormapMixIn, SymbolMixIn,
                ScatterVisualizationMixIn):
    """2D scatter data with settable visualization mode.

    :param parent: The View widget this item belongs to.
    """

    _VISUALIZATION_PROPERTIES = {
        ScatterVisualizationMixIn.Visualization.POINTS:
            ('symbol', 'symbolSize'),
        ScatterVisualizationMixIn.Visualization.LINES:
            ('lineWidth',),
        ScatterVisualizationMixIn.Visualization.SOLID: (),
    }
    """Dict {visualization mode: property names used in this mode}"""

    _SUPPORTED_SCATTER_VISUALIZATION = tuple(_VISUALIZATION_PROPERTIES.keys())
    """Overrides supported Visualizations"""

    def __init__(self, parent=None):
        DataItem3D.__init__(self, parent=parent)
        ColormapMixIn.__init__(self)
        SymbolMixIn.__init__(self)
        ScatterVisualizationMixIn.__init__(self)

        self._heightMap = False
        self._lineWidth = 1.

        self._x = numpy.zeros((0,), dtype=numpy.float32)
        self._y = numpy.zeros((0,), dtype=numpy.float32)
        self._value = numpy.zeros((0,), dtype=numpy.float32)

        self._cachedLinesIndices = None
        self._cachedTrianglesIndices = None

        # Connect scene primitive to mix-in class
        ColormapMixIn._setSceneColormap(self, function.Colormap())

    def _updated(self, event=None):
        """Handle mix-in class updates"""
        if event in (ItemChangedType.SYMBOL, ItemChangedType.SYMBOL_SIZE):
            symbol, size = self._getSceneSymbol()
            for child in self._getScenePrimitive().children:
                if isinstance(child, primitives.Points):
                    child.marker = symbol
                    child.setAttribute('size', size, copy=True)

        elif event is ItemChangedType.VISIBLE:
            # TODO smart update?, need dirty flags
            self._updateScene()

        elif event is ItemChangedType.VISUALIZATION_MODE:
            self._updateScene()

        super(Scatter2D, self)._updated(event)

    def isPropertyEnabled(self, name, visualization=None):
        """Returns true if the property is used with visualization mode.

        :param str name: The name of the property to check, in:
                         'lineWidth', 'symbol', 'symbolSize'
        :param str visualization:
            The visualization mode for which to get the info.
            By default, it is the current visualization mode.
        :return:
        """
        assert name in ('lineWidth', 'symbol', 'symbolSize')
        if visualization is None:
            visualization = self.getVisualization()
        assert visualization in self.supportedVisualizations()
        return name in self._VISUALIZATION_PROPERTIES[visualization]

    def setHeightMap(self, heightMap):
        """Set whether to display the data has a height map or not.

        When displayed as a height map, the data values are used as
        z coordinates.

        :param bool heightMap:
            True to display a height map,
            False to display as 2D data with z=0
        """
        heightMap = bool(heightMap)
        if heightMap != self.isHeightMap():
            self._heightMap = heightMap
            self._updateScene()
            self._updated(Item3DChangedType.HEIGHT_MAP)

    def isHeightMap(self):
        """Returns True if data is displayed as a height map.

        :rtype: bool
        """
        return self._heightMap

    def getLineWidth(self):
        """Return the curve line width in pixels (float)"""
        return self._lineWidth

    def setLineWidth(self, width):
        """Set the width in pixel of the curve line

        See :meth:`getLineWidth`.

        :param float width: Width in pixels
        """
        width = float(width)
        assert width >= 1.
        if width != self._lineWidth:
            self._lineWidth = width
            for child in self._getScenePrimitive().children:
                if hasattr(child, 'lineWidth'):
                    child.lineWidth = width
            self._updated(ItemChangedType.LINE_WIDTH)

    def setData(self, x, y, value, copy=True):
        """Set the data represented by this item.

        Provided arrays must have the same length.

        :param numpy.ndarray x: X coordinates (array-like)
        :param numpy.ndarray y: Y coordinates (array-like)
        :param value: Points value: array-like or single scalar
        :param bool copy:
            True (default) to make a copy of the data,
            False to avoid copy if possible (do not modify the arrays).
        """
        x = numpy.array(
            x, copy=copy, dtype=numpy.float32, order='C').reshape(-1)
        y = numpy.array(
            y, copy=copy, dtype=numpy.float32, order='C').reshape(-1)
        assert len(x) == len(y)

        if isinstance(value, abc.Iterable):
            value = numpy.array(
                value, copy=copy, dtype=numpy.float32, order='C').reshape(-1)
            assert len(value) == len(x)
        else:  # Single scalar
            value = numpy.array((float(value),), dtype=numpy.float32)

        self._x = x
        self._y = y
        self._value = value

        # Reset cache
        self._cachedLinesIndices = None
        self._cachedTrianglesIndices = None

        self._setColormappedData(self.getValueData(copy=False), copy=False)

        self._updateScene()

        self._updated(ItemChangedType.DATA)

    def getData(self, copy=True):
        """Returns data as provided to :meth:`setData`.

        :param bool copy: True to get a copy,
                          False to return internal data (do not modify!)
        :return: (x, y, value)
        """
        return (self.getXData(copy=copy),
                self.getYData(copy=copy),
                self.getValueData(copy=copy))

    def getXData(self, copy=True):
        """Returns X data coordinates.

        :param bool copy: True to get a copy,
                          False to return internal array (do not modify!)
        :return: X coordinates
        :rtype: numpy.ndarray
        """
        return numpy.array(self._x, copy=copy)

    def getYData(self, copy=True):
        """Returns Y data coordinates.

        :param bool copy: True to get a copy,
                          False to return internal array (do not modify!)
        :return: Y coordinates
        :rtype: numpy.ndarray
        """
        return numpy.array(self._y, copy=copy)

    def getValueData(self, copy=True):
        """Returns data values.

        :param bool copy: True to get a copy,
                          False to return internal array (do not modify!)
        :return: data values
        :rtype: numpy.ndarray
        """
        return numpy.array(self._value, copy=copy)

    @deprecated(reason="Consistency with PlotWidget items",
                replacement="getValueData", since_version="0.10.0")
    def getValues(self, copy=True):
        return self.getValueData(copy)

    def _pickPoints(self, context, points, threshold=1., sort='depth'):
        """Perform picking while in 'points' visualization mode

        :param PickContext context: Current picking context
        :param float threshold: Picking threshold in pixel.
            Perform picking in a square of size threshold x threshold.
        :param str sort: How returned indices are sorted:

            - 'index' (default): sort by the value of the indices
            - 'depth':  Sort by the depth of the points from the current
              camera point of view.
        :return: Object holding the results or None
        :rtype: Union[None,PickingResult]
        """
        assert sort in ('index', 'depth')

        rayNdc = context.getPickingSegment(frame='ndc')
        if rayNdc is None:  # No picking outside viewport
            return None

        # Project data to NDC
        primitive = self._getScenePrimitive()
        pointsNdc = primitive.objectToNDCTransform.transformPoints(
            points, perspectiveDivide=True)

        # Perform picking
        distancesNdc = numpy.abs(pointsNdc[:, :2] - rayNdc[0, :2])
        thresholdNdc = threshold / numpy.array(primitive.viewport.size)
        picked = numpy.where(numpy.logical_and(
            numpy.all(distancesNdc < thresholdNdc, axis=1),
            numpy.logical_and(rayNdc[0, 2] <= pointsNdc[:, 2],
                              pointsNdc[:, 2] <= rayNdc[1, 2])))[0]

        if sort == 'depth':
            # Sort picked points from front to back
            picked = picked[numpy.argsort(pointsNdc[picked, 2])]

        if picked.size > 0:
            return PickingResult(self,
                                 positions=points[picked, :3],
                                 indices=picked,
                                 fetchdata=self.getValueData)
        else:
            return None

    def _pickSolid(self, context, points):
        """Perform picking while in 'solid' visualization mode

        :param PickContext context: Current picking context
        """
        if self._cachedTrianglesIndices is None:
            _logger.info("Picking on Scatter2D before rendering")
            return None

        rayObject = context.getPickingSegment(frame=self._getScenePrimitive())
        if rayObject is None:  # No picking outside viewport
            return None
        rayObject = rayObject[:, :3]

        trianglesIndices = self._cachedTrianglesIndices.reshape(-1, 3)
        triangles = points[trianglesIndices, :3]
        selectedIndices, t, barycentric = glu.segmentTrianglesIntersection(
            rayObject, triangles)
        closest = numpy.argmax(barycentric, axis=1)

        indices = trianglesIndices.reshape(-1, 3)[selectedIndices, closest]

        if len(indices) == 0:  # No point is picked
            return None

        # Compute intersection points and get closest data point
        positions = t.reshape(-1, 1) * (rayObject[1] - rayObject[0]) + rayObject[0]

        return PickingResult(self,
                             positions=positions,
                             indices=indices,
                             fetchdata=self.getValueData)

    def _pickFull(self, context):
        """Perform picking in this item at given widget position.

        :param PickContext context: Current picking context
        :return: Object holding the results or None
        :rtype: Union[None,PickingResult]
        """
        xData = self.getXData(copy=False)
        if len(xData) == 0:  # No data in the scatter
            return None

        if self.isHeightMap():
            zData = self.getValueData(copy=False)
        else:
            zData = numpy.zeros_like(xData)

        points = numpy.transpose((xData,
                                  self.getYData(copy=False),
                                  zData,
                                  numpy.ones_like(xData)))

        mode = self.getVisualization()
        if mode is self.Visualization.POINTS:
            # TODO issue with symbol size: using pixel instead of points
            # Get "corrected" symbol size
            _, threshold = self._getSceneSymbol()
            return self._pickPoints(
                context, points, threshold=max(3., threshold))

        elif mode is self.Visualization.LINES:
            # Picking only at point
            return self._pickPoints(context, points, threshold=5.)

        else:  # mode == 'solid'
            return self._pickSolid(context, points)

    def _updateScene(self):
        self._getScenePrimitive().children = []  # Remove previous primitives

        if not self.isVisible():
            return  # Update when visible

        x, y, value = self.getData(copy=False)
        if len(x) == 0:
            return  # Nothing to display

        mode = self.getVisualization()
        heightMap = self.isHeightMap()

        if mode is self.Visualization.POINTS:
            z = value if heightMap else 0.
            symbol, size = self._getSceneSymbol()
            primitive = primitives.Points(
                x=x, y=y, z=z, value=value,
                size=size,
                colormap=self._getSceneColormap())
            primitive.marker = symbol

        else:
            # TODO run delaunay in a thread
            # Compute lines/triangles indices if not cached
            if self._cachedTrianglesIndices is None:
                triangulation = delaunay(x, y)
                if triangulation is None:
                    return None
                self._cachedTrianglesIndices = numpy.ravel(
                    triangulation.simplices.astype(numpy.uint32))

            if (mode is self.Visualization.LINES and
                    self._cachedLinesIndices is None):
                # Compute line indices
                self._cachedLinesIndices = utils.triangleToLineIndices(
                    self._cachedTrianglesIndices, unicity=True)

            if mode is self.Visualization.LINES:
                indices = self._cachedLinesIndices
                renderMode = 'lines'
            else:
                indices = self._cachedTrianglesIndices
                renderMode = 'triangles'

            # TODO supports x, y instead of copy
            if heightMap:
                if len(value) == 1:
                    value = numpy.ones_like(x) * value
                coordinates = numpy.array((x, y, value), dtype=numpy.float32).T
            else:
                coordinates = numpy.array((x, y), dtype=numpy.float32).T

            # TODO option to enable/disable light, cache normals
            # TODO smooth surface
            if mode is self.Visualization.SOLID:
                if heightMap:
                    coordinates = coordinates[indices]
                    if len(value) > 1:
                        value = value[indices]
                    triangleNormals = utils.trianglesNormal(coordinates)
                    normal = numpy.empty((len(triangleNormals) * 3, 3),
                                         dtype=numpy.float32)
                    normal[0::3, :] = triangleNormals
                    normal[1::3, :] = triangleNormals
                    normal[2::3, :] = triangleNormals
                    indices = None
                else:
                    normal = (0., 0., 1.)
            else:
                normal = None

            primitive = primitives.ColormapMesh3D(
                coordinates,
                value.reshape(-1, 1),  # Makes it a 2D array
                normal=normal,
                colormap=self._getSceneColormap(),
                indices=indices,
                mode=renderMode)
            primitive.lineWidth = self.getLineWidth()
            primitive.lineSmooth = False

        self._getScenePrimitive().children = [primitive]



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Basic test of Qt bindings wrapper."""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "05/12/2016"


import os.path
import unittest

from silx.test.utils import temp_dir
from silx.gui.utils.testutils import TestCaseQt

from silx.gui import qt
try:
    from silx.gui.qt import inspect as qt_inspect
except ImportError:
    qt_inspect = None


class TestQtWrapper(unittest.TestCase):
    """Minimalistic test to check that Qt has been loaded."""

    def testQObject(self):
        """Test that QObject is there."""
        obj = qt.QObject()
        self.assertTrue(obj is not None)


class TestLoadUi(TestCaseQt):
    """Test loadUi function"""

    TEST_UI = """<?xml version="1.0" encoding="UTF-8"?>
    <ui version="4.0">
     <class>MainWindow</class>
     <widget class="QMainWindow" name="MainWindow">
      <property name="geometry">
       <rect>
        <x>0</x>
        <y>0</y>
        <width>293</width>
        <height>296</height>
       </rect>
      </property>
      <property name="windowTitle">
       <string>Test loadUi</string>
      </property>
      <widget class="QWidget" name="centralwidget">
       <widget class="QPushButton" name="pushButton">
        <property name="geometry">
         <rect>
          <x>10</x>
          <y>10</y>
          <width>89</width>
          <height>27</height>
         </rect>
        </property>
        <property name="text">
         <string>Button 1</string>
        </property>
       </widget>
       <widget class="QPushButton" name="pushButton_2">
        <property name="geometry">
         <rect>
          <x>10</x>
          <y>50</y>
          <width>89</width>
          <height>27</height>
         </rect>
        </property>
        <property name="text">
         <string>Button 2</string>
        </property>
       </widget>
       <widget class="Line" name="line">
        <property name="geometry">
         <rect>
          <x>10</x>
          <y>90</y>
          <width>118</width>
          <height>3</height>
         </rect>
        </property>
        <property name="orientation">
         <enum>Qt::Horizontal</enum>
        </property>
       </widget>
       <widget class="Line" name="line_2">
        <property name="geometry">
         <rect>
          <x>150</x>
          <y>20</y>
          <width>3</width>
          <height>61</height>
         </rect>
        </property>
        <property name="orientation">
         <enum>Qt::Vertical</enum>
        </property>
       </widget>
      </widget>
      <widget class="QMenuBar" name="menubar">
       <property name="geometry">
        <rect>
         <x>0</x>
         <y>0</y>
         <width>293</width>
         <height>25</height>
        </rect>
       </property>
      </widget>
      <widget class="QStatusBar" name="statusbar"/>
     </widget>
     <resources/>
     <connections/>
    </ui>
    """

    @unittest.skipIf(qt.BINDING == "PySide", "Not fully working with PySide")
    def testLoadUi(self):
        """Create a QMainWindow from an ui file"""
        with temp_dir() as tmp:
            uifile = os.path.join(tmp, "test.ui")

            # write file
            with open(uifile, mode='w') as f:
                f.write(self.TEST_UI)

            class TestMainWindow(qt.QMainWindow):
                def __init__(self, parent=None):
                    super(TestMainWindow, self).__init__(parent)
                    qt.loadUi(uifile, self)

            testMainWindow = TestMainWindow()
            testMainWindow.show()
            self.qWaitForWindowExposed(testMainWindow)

            testMainWindow.setAttribute(qt.Qt.WA_DeleteOnClose)
            testMainWindow.close()


class TestQtInspect(unittest.TestCase):
    """Test functions of silx.gui.qt.inspect module"""

    # shiboken module is not always available
    @unittest.skipIf(qt.BINDING == 'PySide' and qt_inspect is None,
                     reason="shiboken module not available")
    def test(self):
        """Test functions of silx.gui.qt.inspect module"""
        self.assertIsNotNone(qt_inspect)

        parent = qt.QObject()

        self.assertTrue(qt_inspect.isValid(parent))
        self.assertTrue(qt_inspect.createdByPython(parent))
        self.assertTrue(qt_inspect.ownedByPython(parent))

        obj = qt.QObject(parent)

        self.assertTrue(qt_inspect.isValid(obj))
        self.assertTrue(qt_inspect.createdByPython(obj))
        self.assertFalse(qt_inspect.ownedByPython(obj))

        del parent
        self.assertFalse(qt_inspect.isValid(obj))


def suite():
    test_suite = unittest.TestSuite()
    for TestCaseCls in (TestQtWrapper, TestLoadUi, TestQtInspect):
        test_suite.addTest(
            unittest.defaultTestLoader.loadTestsFromTestCase(TestCaseCls))
    return test_suite


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Image stack view with data prefetch capabilty."""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "04/03/2019"


from silx.gui import icons, qt
from silx.gui.plot import Plot2D
from silx.gui.utils import concurrent
from silx.io.url import DataUrl
from silx.io.utils import get_data
from collections import OrderedDict
from silx.gui.widgets.FrameBrowser import HorizontalSliderWithBrowser
import time
import threading
import typing
import logging

_logger = logging.getLogger(__file__)


class _PlotWithWaitingLabel(qt.QWidget):
    """Image plot widget with an overlay 'waiting' status.
    """

    class AnimationThread(threading.Thread):
        def __init__(self, label):
            self.running = True
            self._label = label
            self.animated_icon = icons.getWaitIcon()
            self.animated_icon.register(self._label)
            super(_PlotWithWaitingLabel.AnimationThread, self).__init__()

        def run(self):
            while self.running:
                time.sleep(0.05)
                icon = self.animated_icon.currentIcon()
                self.future_result = concurrent.submitToQtMainThread(
                    self._label.setPixmap, icon.pixmap(30, state=qt.QIcon.On))

        def stop(self):
            """Stop the update thread"""
            self.animated_icon.unregister(self._label)
            self.running = False
            self.join(2)

    def __init__(self, parent):
        super(_PlotWithWaitingLabel, self).__init__(parent=parent)
        layout = qt.QStackedLayout(self)
        layout.setStackingMode(qt.QStackedLayout.StackAll)

        self._waiting_label = qt.QLabel(parent=self)
        self._waiting_label.setAlignment(qt.Qt.AlignHCenter | qt.Qt.AlignVCenter)
        layout.addWidget(self._waiting_label)

        self._plot = Plot2D(parent=self)
        layout.addWidget(self._plot)

        self.updateThread = _PlotWithWaitingLabel.AnimationThread(self._waiting_label)
        self.updateThread.start()

    def close(self) -> bool:
        super(_PlotWithWaitingLabel, self).close()
        self.updateThread.stop()

    def setWaiting(self, activate=True):
        if activate is True:
            self._plot.clear()
            self._waiting_label.show()
        else:
            self._waiting_label.hide()

    def setData(self, data):
        self.setWaiting(activate=False)
        self._plot.addImage(data=data)

    def clear(self):
        self._plot.clear()
        self.setWaiting(False)

    def getPlotWidget(self):
        return self._plot


class _HorizontalSlider(HorizontalSliderWithBrowser):

    sigCurrentUrlIndexChanged = qt.Signal(int)

    def __init__(self, parent):
        super(_HorizontalSlider, self).__init__(parent=parent)
        #  connect signal / slot
        self.valueChanged.connect(self._urlChanged)

    def setUrlIndex(self, index):
        self.setValue(index)
        self.sigCurrentUrlIndexChanged.emit(index)

    def _urlChanged(self, value):
        self.sigCurrentUrlIndexChanged.emit(value)


class UrlList(qt.QWidget):
    """List of URLs the user to select an URL"""

    sigCurrentUrlChanged = qt.Signal(str)
    """Signal emitted when the active/current url change"""

    def __init__(self, parent=None):
        super(UrlList, self).__init__(parent)
        self.setLayout(qt.QVBoxLayout())
        self.layout().setSpacing(0)
        self.layout().setContentsMargins(0, 0, 0, 0)
        self._listWidget = qt.QListWidget(parent=self)
        self.layout().addWidget(self._listWidget)

        # connect signal / Slot
        self._listWidget.currentItemChanged.connect(self._notifyCurrentUrlChanged)

        # expose API
        self.currentItem = self._listWidget.currentItem

    def setUrls(self, urls: list) -> None:
        url_names = []
        [url_names.append(url.path()) for url in urls]
        self._listWidget.addItems(url_names)

    def _notifyCurrentUrlChanged(self, current, previous):
        if current is None:
            pass
        else:
            self.sigCurrentUrlChanged.emit(current.text())

    def setUrl(self, url: DataUrl) -> None:
        assert isinstance(url, DataUrl)
        sel_items = self._listWidget.findItems(url.path(), qt.Qt.MatchExactly)
        if sel_items is None:
            _logger.warning(url.path(), ' is not registered in the list.')
        else:
            assert len(sel_items) == 1
            item = sel_items[0]
            self._listWidget.setCurrentItem(item)
            self.sigCurrentUrlChanged.emit(item.text())

    def clear(self):
        self._listWidget.clear()


class _ToggleableUrlSelectionTable(qt.QWidget):

    _BUTTON_ICON = qt.QStyle.SP_ToolBarHorizontalExtensionButton  # noqa

    sigCurrentUrlChanged = qt.Signal(str)
    """Signal emitted when the active/current url change"""

    def __init__(self, parent=None) -> None:
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())
        self._toggleButton = qt.QPushButton(parent=self)
        self.layout().addWidget(self._toggleButton, 0, 2, 1, 1)
        self._toggleButton.setSizePolicy(qt.QSizePolicy.Fixed,
                                         qt.QSizePolicy.Fixed)

        self._urlsTable = UrlList(parent=self)
        self.layout().addWidget(self._urlsTable, 1, 1, 1, 2)

        # set up
        self._setButtonIcon(show=True)

        # Signal / slot connection
        self._toggleButton.clicked.connect(self.toggleUrlSelectionTable)
        self._urlsTable.sigCurrentUrlChanged.connect(self._propagateSignal)

        # expose API
        self.setUrls = self._urlsTable.setUrls
        self.setUrl = self._urlsTable.setUrl
        self.currentItem = self._urlsTable.currentItem

    def toggleUrlSelectionTable(self):
        visible = not self.urlSelectionTableIsVisible()
        self._setButtonIcon(show=visible)
        self._urlsTable.setVisible(visible)

    def _setButtonIcon(self, show):
        style = qt.QApplication.instance().style()
        # return a QIcon
        icon = style.standardIcon(self._BUTTON_ICON)
        if show is False:
            pixmap = icon.pixmap(32, 32).transformed(qt.QTransform().scale(-1, 1))
            icon = qt.QIcon(pixmap)
        self._toggleButton.setIcon(icon)

    def urlSelectionTableIsVisible(self):
        return self._urlsTable.isVisible()

    def _propagateSignal(self, url):
        self.sigCurrentUrlChanged.emit(url)

    def clear(self):
        self._urlsTable.clear()


class UrlLoader(qt.QThread):
    """
    Thread use to load DataUrl
    """
    def __init__(self, parent, url):
        super(UrlLoader, self).__init__(parent=parent)
        assert isinstance(url, DataUrl)
        self.url = url
        self.data = None

    def run(self):
        try:
            self.data = get_data(self.url)
        except IOError:
            self.data = None


class ImageStack(qt.QMainWindow):
    """Widget loading on the fly images contained the given urls.

    It prefetches images close to the displayed one.
    """

    N_PRELOAD = 10

    sigLoaded = qt.Signal(str)
    """Signal emitted when new data is available"""

    sigCurrentUrlChanged = qt.Signal(str)
    """Signal emitted when the current url change"""

    def __init__(self, parent=None) -> None:
        super(ImageStack, self).__init__(parent)
        self.__n_prefetch = ImageStack.N_PRELOAD
        self._loadingThreads = []
        self.setWindowFlags(qt.Qt.Widget)
        self._current_url = None
        self._url_loader = UrlLoader
        "class to instantiate for loading urls"

        # main widget
        self._plot = _PlotWithWaitingLabel(parent=self)
        self._plot.setAttribute(qt.Qt.WA_DeleteOnClose, True)
        self.setWindowTitle("Image stack")
        self.setCentralWidget(self._plot)

        # dock widget: url table
        self._tableDockWidget = qt.QDockWidget(parent=self)
        self._urlsTable = _ToggleableUrlSelectionTable(parent=self)
        self._tableDockWidget.setWidget(self._urlsTable)
        self._tableDockWidget.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self._tableDockWidget)
        # dock widget: qslider
        self._sliderDockWidget = qt.QDockWidget(parent=self)
        self._slider = _HorizontalSlider(parent=self)
        self._sliderDockWidget.setWidget(self._slider)
        self.addDockWidget(qt.Qt.BottomDockWidgetArea, self._sliderDockWidget)
        self._sliderDockWidget.setFeatures(qt.QDockWidget.DockWidgetMovable)

        self.reset()

        # connect signal / slot
        self._urlsTable.sigCurrentUrlChanged.connect(self.setCurrentUrl)
        self._slider.sigCurrentUrlIndexChanged.connect(self.setCurrentUrlIndex)

    def close(self) -> bool:
        self._freeLoadingThreads()
        self._plot.close()
        super(ImageStack, self).close()

    def setUrlLoaderClass(self, urlLoader: typing.Type[UrlLoader]) -> None:
        """

        :param urlLoader: define the class to call for loading urls.
                          warning: this should be a class object and not a
                          class instance.
        """
        assert isinstance(urlLoader, type(UrlLoader))
        self._url_loader = urlLoader

    def getUrlLoaderClass(self):
        """

        :return: class to instantiate for loading urls
        :rtype: typing.Type[UrlLoader]
        """
        return self._url_loader

    def _freeLoadingThreads(self):
        for thread in self._loadingThreads:
            thread.blockSignals(True)
            thread.wait(5)
        self._loadingThreads.clear()

    def getPlotWidget(self) -> Plot2D:
        """
        Returns the PlotWidget contained in this window

        :return: PlotWidget contained in this window
        :rtype: Plot2D
        """
        return self._plot.getPlotWidget()

    def reset(self) -> None:
        """Clear the plot and remove any link to url"""
        self._freeLoadingThreads()
        self._urls = None
        self._urlIndexes = None
        self._urlData = OrderedDict({})
        self._current_url = None
        self._plot.clear()
        self._urlsTable.clear()
        self._slider.setMaximum(-1)

    def _preFetch(self, urls: list) -> None:
        """Pre-fetch the given urls if necessary

        :param urls: list of DataUrl to prefetch
        :type: list
        """
        for url in urls:
            if url.path() not in self._urlData:
                self._load(url)

    def _load(self, url):
        """
        Launch background load of a DataUrl

        :param url:
        :type: DataUrl
        """
        assert isinstance(url, DataUrl)
        url_path = url.path()
        assert url_path in self._urlIndexes
        loader = self._url_loader(parent=self, url=url)
        loader.finished.connect(self._urlLoaded, qt.Qt.QueuedConnection)
        self._loadingThreads.append(loader)
        loader.start()

    def _urlLoaded(self) -> None:
        """

        :param url: restul of DataUrl.path() function
        :return:
        """
        sender = self.sender()
        assert isinstance(sender, UrlLoader)
        url = sender.url.path()
        if url in self._urlIndexes:
            self._urlData[url] = sender.data
            if self.getCurrentUrl().path() == url:
                self._plot.setData(self._urlData[url])
            if sender in self._loadingThreads:
                self._loadingThreads.remove(sender)
            self.sigLoaded.emit(url)

    def setNPrefetch(self, n: int) -> None:
        """
        Define the number of url to prefetch around

        :param int n: number of url to prefetch on left and right sides.
                      In total n*2 DataUrl will be prefetch
        """
        self.__n_prefetch = n
        current_url = self.getCurrentUrl()
        if current_url is not None:
            self.setCurrentUrl(current_url)

    def getNPrefetch(self) -> int:
        """

        :return: number of url to prefetch on left and right sides. In total
                 will load 2* NPrefetch DataUrls
        """
        return self.__n_prefetch

    def setUrls(self, urls: list) -> None:
        """list of urls within an index. Warning: urls should contain an image
        compatible with the silx.gui.plot.Plot class

        :param urls: urls we want to set in the stack. Key is the index
                     (position in the stack), value is the DataUrl
        :type: list
        """
        def createUrlIndexes():
            indexes = OrderedDict()
            for index, url in enumerate(urls):
                indexes[index] = url
            return indexes

        urls_with_indexes = createUrlIndexes()
        urlsToIndex = self._urlsToIndex(urls_with_indexes)
        self.reset()
        self._urls = urls_with_indexes
        self._urlIndexes = urlsToIndex

        old_url_table = self._urlsTable.blockSignals(True)
        self._urlsTable.setUrls(urls=list(self._urls.values()))
        self._urlsTable.blockSignals(old_url_table)

        old_slider = self._slider.blockSignals(True)
        self._slider.setMinimum(0)
        self._slider.setMaximum(len(self._urls) - 1)
        self._slider.blockSignals(old_slider)

        if self.getCurrentUrl() in self._urls:
            self.setCurrentUrl(self.getCurrentUrl())
        else:
            if len(self._urls.keys()) > 0:
                first_url = self._urls[list(self._urls.keys())[0]]
                self.setCurrentUrl(first_url)

    def getUrls(self) -> tuple:
        """

        :return: tuple of urls
        :rtype: tuple
        """
        return tuple(self._urlIndexes.keys())

    def _getNextUrl(self, url: DataUrl) -> typing.Union[None, DataUrl]:
        """
        return the next url in the stack

        :param url: url for which we want the next url
        :type: DataUrl
        :return: next url in the stack or None if `url` is the last one
        :rtype: Union[None, DataUrl]
        """
        assert isinstance(url, DataUrl)
        if self._urls is None:
            return None
        else:
            index = self._urlIndexes[url.path()]
            indexes = list(self._urls.keys())
            res = list(filter(lambda x: x > index, indexes))
            if len(res) == 0:
                return None
            else:
                return self._urls[res[0]]

    def _getPreviousUrl(self, url: DataUrl) -> typing.Union[None, DataUrl]:
        """
        return the previous url in the stack

        :param url: url for which we want the previous url
        :type: DataUrl
        :return: next url in the stack or None if `url` is the last one
        :rtype: Union[None, DataUrl]
        """
        if self._urls is None:
            return None
        else:
            index = self._urlIndexes[url.path()]
            indexes = list(self._urls.keys())
            res = list(filter(lambda x: x < index, indexes))
            if len(res) == 0:
                return None
            else:
                return self._urls[res[-1]]

    def _getNNextUrls(self, n: int, url: DataUrl) -> list:
        """
        Deduce the next urls in the stack after `url`

        :param n: the number of url store after `url`
        :type: int
        :param url: url for which we want n next url
        :type: DataUrl
        :return: list of next urls.
        :rtype: list
        """
        res = []
        next_free = self._getNextUrl(url=url)
        while len(res) < n and next_free is not None:
            assert isinstance(next_free, DataUrl)
            res.append(next_free)
            next_free = self._getNextUrl(res[-1])
        return res

    def _getNPreviousUrls(self, n: int, url: DataUrl):
        """
        Deduce the previous urls in the stack after `url`

        :param n: the number of url store after `url`
        :type: int
        :param url: url for which we want n previous url
        :type: DataUrl
        :return: list of previous urls.
        :rtype: list
        """
        res = []
        next_free = self._getPreviousUrl(url=url)
        while len(res) < n and next_free is not None:
            res.insert(0, next_free)
            next_free = self._getPreviousUrl(res[0])
        return res

    def setCurrentUrlIndex(self, index: int):
        """
        Define the url to be displayed

        :param index: url to be displayed
        :type: int
        """
        if index < 0:
            return
        if self._urls is None:
            return
        elif index >= len(self._urls):
            raise ValueError('requested index out of bounds')
        else:
            return self.setCurrentUrl(self._urls[index])

    def setCurrentUrl(self, url: typing.Union[DataUrl, str]) -> None:
        """
        Define the url to be displayed

        :param url: url to be displayed
        :type: DataUrl
        """
        assert isinstance(url, (DataUrl, str))
        if isinstance(url, str):
            url = DataUrl(path=url)
        if url != self._current_url:
            self._current_url = url
            self.sigCurrentUrlChanged.emit(url.path())

        old_url_table = self._urlsTable.blockSignals(True)
        old_slider = self._slider.blockSignals(True)

        self._urlsTable.setUrl(url)
        self._slider.setUrlIndex(self._urlIndexes[url.path()])
        if self._current_url is None:
            self._plot.clear()
        else:
            if self._current_url.path() in self._urlData:
                self._plot.setData(self._urlData[url.path()])
            else:
                self._load(url)
                self._notifyLoading()
            self._preFetch(self._getNNextUrls(self.__n_prefetch, url))
            self._preFetch(self._getNPreviousUrls(self.__n_prefetch, url))
        self._urlsTable.blockSignals(old_url_table)
        self._slider.blockSignals(old_slider)

    def getCurrentUrl(self) -> typing.Union[None, DataUrl]:
        """

        :return: url currently displayed
        :rtype: Union[None, DataUrl]
        """
        return self._current_url

    def getCurrentUrlIndex(self) -> typing.Union[None, int]:
        """

        :return: index of the url currently displayed
        :rtype: Union[None, int]
        """
        if self._current_url is None:
            return None
        else:
            return self._urlIndexes[self._current_url.path()]

    @staticmethod
    def _urlsToIndex(urls):
        """util, return a dictionary with url as key and index as value"""
        res = {}
        for index, url in urls.items():
            res[url.path()] = index
        return res

    def _notifyLoading(self):
        """display a simple image of loading..."""
        self._plot.setWaiting(activate=True)




# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2004-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Miscellaneous utility functions for the Plot"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "21/03/2017"


import numpy

from .panzoom import FLOAT32_SAFE_MIN, FLOAT32_MINPOS, FLOAT32_SAFE_MAX
from .panzoom import applyZoomToPlot, applyPan


def addMarginsToLimits(margins, isXLog, isYLog,
                       xMin, xMax, yMin, yMax, y2Min=None, y2Max=None):
    """Returns updated limits by extending them with margins.

    :param margins: The ratio of the margins to add or None for no margins.
    :type margins: A 4-tuple of floats as
                   (xMinMargin, xMaxMargin, yMinMargin, yMaxMargin)

    :return: The updated limits
    :rtype: tuple of 4 or 6 floats: Either (xMin, xMax, yMin, yMax) or
            (xMin, xMax, yMin, yMax, y2Min, y2Max) if y2Min and y2Max
            are provided.
    """
    if margins is not None:
        xMinMargin, xMaxMargin, yMinMargin, yMaxMargin = margins

        if not isXLog:
            xRange = xMax - xMin
            xMin -= xMinMargin * xRange
            xMax += xMaxMargin * xRange

        elif xMin > 0. and xMax > 0.:  # Log scale
            # Do not apply margins if limits < 0
            xMinLog, xMaxLog = numpy.log10(xMin), numpy.log10(xMax)
            xRangeLog = xMaxLog - xMinLog
            xMin = pow(10., xMinLog - xMinMargin * xRangeLog)
            xMax = pow(10., xMaxLog + xMaxMargin * xRangeLog)

        if not isYLog:
            yRange = yMax - yMin
            yMin -= yMinMargin * yRange
            yMax += yMaxMargin * yRange
        elif yMin > 0. and yMax > 0.:  # Log scale
            # Do not apply margins if limits < 0
            yMinLog, yMaxLog = numpy.log10(yMin), numpy.log10(yMax)
            yRangeLog = yMaxLog - yMinLog
            yMin = pow(10., yMinLog - yMinMargin * yRangeLog)
            yMax = pow(10., yMaxLog + yMaxMargin * yRangeLog)

        if y2Min is not None and y2Max is not None:
            if not isYLog:
                yRange = y2Max - y2Min
                y2Min -= yMinMargin * yRange
                y2Max += yMaxMargin * yRange
            elif y2Min > 0. and y2Max > 0.:  # Log scale
                # Do not apply margins if limits < 0
                yMinLog, yMaxLog = numpy.log10(y2Min), numpy.log10(y2Max)
                yRangeLog = yMaxLog - yMinLog
                y2Min = pow(10., yMinLog - yMinMargin * yRangeLog)
                y2Max = pow(10., yMaxLog + yMaxMargin * yRangeLog)

    if y2Min is None or y2Max is None:
        return xMin, xMax, yMin, yMax
    else:
        return xMin, xMax, yMin, yMax, y2Min, y2Max




# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This package provides classes that describes :class:`.PlotWidget` content.

Instances of those classes are returned by :class:`.PlotWidget` methods that give
access to its content such as :meth:`.PlotWidget.getCurve`, :meth:`.PlotWidget.getImage`.
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "22/06/2017"

from .core import (Item, DataItem,  # noqa
                   LabelsMixIn, DraggableMixIn, ColormapMixIn,  # noqa
                   SymbolMixIn, ColorMixIn, YAxisMixIn, FillMixIn,  # noqa
                   AlphaMixIn, LineMixIn, ScatterVisualizationMixIn,  # noqa
                   ComplexMixIn, ItemChangedType, PointsBase)  # noqa
from .complex import ImageComplexData  # noqa
from .curve import Curve, CurveStyle  # noqa
from .histogram import Histogram  # noqa
from .image import ImageBase, ImageData, ImageRgba, ImageStack, MaskImageData  # noqa
from .shape import Shape, BoundingRect, XAxisExtent, YAxisExtent  # noqa
from .scatter import Scatter  # noqa
from .marker import MarkerBase, Marker, XMarker, YMarker  # noqa
from .axis import Axis, XAxis, YAxis, YRightAxis

DATA_ITEMS = (ImageComplexData, Curve, Histogram, ImageBase, Scatter,
              BoundingRect, XAxisExtent, YAxisExtent)
"""Classes of items representing data and to consider to compute data bounds.
"""



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2004-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Matplotlib Plot backend."""

from __future__ import division

__authors__ = ["V.A. Sole", "T. Vincent, H. Payno"]
__license__ = "MIT"
__date__ = "21/12/2018"


import logging
import datetime as dt
from typing import Tuple
import numpy

from pkg_resources import parse_version as _parse_version


_logger = logging.getLogger(__name__)


from ... import qt

# First of all init matplotlib and set its backend
from ...utils.matplotlib import FigureCanvasQTAgg
import matplotlib
from matplotlib.container import Container
from matplotlib.figure import Figure
from matplotlib.patches import Rectangle, Polygon
from matplotlib.image import AxesImage
from matplotlib.backend_bases import MouseEvent
from matplotlib.lines import Line2D
from matplotlib.text import Text
from matplotlib.collections import PathCollection, LineCollection
from matplotlib.ticker import Formatter, ScalarFormatter, Locator
from matplotlib.tri import Triangulation
from matplotlib.collections import TriMesh
from matplotlib import path as mpath

from . import BackendBase
from .. import items
from .._utils import FLOAT32_MINPOS
from .._utils.dtime_ticklayout import calcTicks, bestFormatString, timestamp

_PATCH_LINESTYLE = {
    "-": 'solid',
    "--": 'dashed',
    '-.': 'dashdot',
    ':': 'dotted',
    '': "solid",
    None: "solid",
}
"""Patches do not uses the same matplotlib syntax"""

_MARKER_PATHS = {}
"""Store cached extra marker paths"""

_SPECIAL_MARKERS = {
    'tickleft': 0,
    'tickright': 1,
    'tickup': 2,
    'tickdown': 3,
    'caretleft': 4,
    'caretright': 5,
    'caretup': 6,
    'caretdown': 7,
}


def normalize_linestyle(linestyle):
    """Normalize known old-style linestyle, else return the provided value."""
    return _PATCH_LINESTYLE.get(linestyle, linestyle)

def get_path_from_symbol(symbol):
    """Get the path representation of a symbol, else None if
    it is not provided.

    :param str symbol: Symbol description used by silx
    :rtype: Union[None,matplotlib.path.Path]
    """
    if symbol == u'\u2665':
        path = _MARKER_PATHS.get(symbol, None)
        if path is not None:
            return path
        vertices = numpy.array([
            [0,-99],
            [31,-73], [47,-55], [55,-46],
            [63,-37], [94,-2], [94,33],
            [94,69], [71,89], [47,89],
            [24,89], [8,74], [0,58],
            [-8,74], [-24,89], [-47,89],
            [-71,89], [-94,69], [-94,33],
            [-94,-2], [-63,-37], [-55,-46],
            [-47,-55], [-31,-73], [0,-99],
            [0,-99]])
        codes = [mpath.Path.CURVE4] * len(vertices)
        codes[0] = mpath.Path.MOVETO
        codes[-1] = mpath.Path.CLOSEPOLY
        path = mpath.Path(vertices, codes)
        _MARKER_PATHS[symbol] = path
        return path
    return None

class NiceDateLocator(Locator):
    """
    Matplotlib Locator that uses Nice Numbers algorithm (adapted to dates)
    to find the tick locations. This results in the same number behaviour
    as when using the silx Open GL backend.

    Expects the data to be posix timestampes (i.e. seconds since 1970)
    """
    def __init__(self, numTicks=5, tz=None):
        """
        :param numTicks: target number of ticks
        :param datetime.tzinfo tz: optional time zone. None is local time.
        """
        super(NiceDateLocator, self).__init__()
        self.numTicks = numTicks

        self._spacing = None
        self._unit = None
        self.tz = tz

    @property
    def spacing(self):
        """ The current spacing. Will be updated when new tick value are made"""
        return self._spacing

    @property
    def unit(self):
        """ The current DtUnit. Will be updated when new tick value are made"""
        return self._unit

    def __call__(self):
        """Return the locations of the ticks"""
        vmin, vmax = self.axis.get_view_interval()
        return self.tick_values(vmin, vmax)

    def tick_values(self, vmin, vmax):
        """ Calculates tick values
        """
        if vmax < vmin:
            vmin, vmax = vmax, vmin

        # vmin and vmax should be timestamps (i.e. seconds since 1 Jan 1970)
        dtMin = dt.datetime.fromtimestamp(vmin, tz=self.tz)
        dtMax = dt.datetime.fromtimestamp(vmax, tz=self.tz)
        dtTicks, self._spacing, self._unit = \
            calcTicks(dtMin, dtMax, self.numTicks)

        # Convert datetime back to time stamps.
        ticks = [timestamp(dtTick) for dtTick in dtTicks]
        return ticks


class NiceAutoDateFormatter(Formatter):
    """
    Matplotlib FuncFormatter that is linked to a NiceDateLocator and gives the
    best possible formats given the locators current spacing an date unit.
    """

    def __init__(self, locator, tz=None):
        """
        :param niceDateLocator: a NiceDateLocator object
        :param datetime.tzinfo tz: optional time zone. None is local time.
        """
        super(NiceAutoDateFormatter, self).__init__()
        self.locator = locator
        self.tz = tz

    @property
    def formatString(self):
        if self.locator.spacing is None or self.locator.unit is None:
            # Locator has no spacing or units yet. Return elaborate fmtString
            return "Y-%m-%d %H:%M:%S"
        else:
            return bestFormatString(self.locator.spacing, self.locator.unit)

    def __call__(self, x, pos=None):
        """Return the format for tick val *x* at position *pos*
           Expects x to be a POSIX timestamp (seconds since 1 Jan 1970)
        """
        dateTime = dt.datetime.fromtimestamp(x, tz=self.tz)
        tickStr = dateTime.strftime(self.formatString)
        return tickStr


class _PickableContainer(Container):
    """Artists container with a :meth:`contains` method"""

    def __init__(self, *args, **kwargs):
        Container.__init__(self, *args, **kwargs)
        self.__zorder = None

    @property
    def axes(self):
        """Mimin Artist.axes"""
        for child in self.get_children():
            if hasattr(child, 'axes'):
                return child.axes
        return None

    def draw(self, *args, **kwargs):
        """artist-like draw to broadcast draw to children"""
        for child in self.get_children():
            child.draw(*args, **kwargs)

    def get_zorder(self):
        """Mimic Artist.get_zorder"""
        return self.__zorder

    def set_zorder(self, z):
        """Mimic Artist.set_zorder to broadcast to children"""
        if z != self.__zorder:
            self.__zorder = z
            for child in self.get_children():
                child.set_zorder(z)

    def contains(self, mouseevent):
        """Mimic Artist.contains, and call it on all children.

        :param mouseevent:
        :return: Picking status and associated information as a dict
        :rtype: (bool,dict)
        """
        # Goes through children from front to back and return first picked one.
        for child in reversed(self.get_children()):
            picked, info = child.contains(mouseevent)
            if picked:
                return picked, info
        return False, {}


class _TextWithOffset(Text):
    """Text object which can be displayed at a specific position
    of the plot, but with a pixel offset"""

    def __init__(self, *args, **kwargs):
        Text.__init__(self, *args, **kwargs)
        self.pixel_offset = (0, 0)
        self.__cache = None

    def draw(self, renderer):
        self.__cache = None
        return Text.draw(self, renderer)

    def __get_xy(self):
        if self.__cache is not None:
            return self.__cache

        align = self.get_horizontalalignment()
        if align == "left":
            xoffset = self.pixel_offset[0]
        elif align == "right":
            xoffset = -self.pixel_offset[0]
        else:
            xoffset = 0

        align = self.get_verticalalignment()
        if align == "top":
            yoffset = -self.pixel_offset[1]
        elif align == "bottom":
            yoffset = self.pixel_offset[1]
        else:
            yoffset = 0

        trans = self.get_transform()
        invtrans = self.get_transform().inverted()

        x = super(_TextWithOffset, self).convert_xunits(self._x)
        y = super(_TextWithOffset, self).convert_xunits(self._y)
        pos = x, y
        proj = trans.transform_point(pos)
        proj = proj + numpy.array((xoffset, yoffset))
        pos = invtrans.transform_point(proj)
        self.__cache = pos
        return pos

    def convert_xunits(self, x):
        """Return the pixel position of the annotated point."""
        return self.__get_xy()[0]

    def convert_yunits(self, y):
        """Return the pixel position of the annotated point."""
        return self.__get_xy()[1]


class _MarkerContainer(_PickableContainer):
    """Marker artists container supporting draw/remove and text position update

    :param artists:
        Iterable with either one Line2D or a Line2D and a Text.
        The use of an iterable if enforced by Container being
        a subclass of tuple that defines a specific __new__.
    :param x: X coordinate of the marker (None for horizontal lines)
    :param y: Y coordinate of the marker (None for vertical lines)
    """

    def __init__(self, artists, symbol, x, y, yAxis):
        self.line = artists[0]
        self.text = artists[1] if len(artists) > 1 else None
        self.symbol = symbol
        self.x = x
        self.y = y
        self.yAxis = yAxis

        _PickableContainer.__init__(self, artists)

    def draw(self, *args, **kwargs):
        """artist-like draw to broadcast draw to line and text"""
        self.line.draw(*args, **kwargs)
        if self.text is not None:
            self.text.draw(*args, **kwargs)

    def updateMarkerText(self, xmin, xmax, ymin, ymax, yinverted):
        """Update marker text position and visibility according to plot limits

        :param xmin: X axis lower limit
        :param xmax: X axis upper limit
        :param ymin: Y axis lower limit
        :param ymax: Y axis upper limit
        :param yinverted: True if the y axis is inverted
        """
        if self.text is not None:
            visible = ((self.x is None or xmin <= self.x <= xmax) and
                       (self.y is None or ymin <= self.y <= ymax))
            self.text.set_visible(visible)

            if self.x is not None and self.y is not None:
                if self.symbol is None:
                    valign = 'baseline'
                else:
                    if yinverted:
                        valign = 'bottom'
                    else:
                        valign = 'top'
                self.text.set_verticalalignment(valign)

            elif self.y is None:  # vertical line
                # Always display it on top
                center = (ymax + ymin) * 0.5
                pos = (ymax - ymin) * 0.5 * 0.99
                if yinverted:
                    pos = -pos
                self.text.set_y(center + pos)

            elif self.x is None:  # Horizontal line
                delta = abs(xmax - xmin)
                if xmin > xmax:
                    xmax = xmin
                xmax -= 0.005 * delta
                self.text.set_x(xmax)

    def contains(self, mouseevent):
        """Mimic Artist.contains, and call it on the line Artist.

        :param mouseevent:
        :return: Picking status and associated information as a dict
        :rtype: (bool,dict)
        """
        return self.line.contains(mouseevent)


class _DoubleColoredLinePatch(matplotlib.patches.Patch):
    """Matplotlib patch to display any patch using double color."""

    def __init__(self, patch):
        super(_DoubleColoredLinePatch, self).__init__()
        self.__patch = patch
        self.linebgcolor = None

    def __getattr__(self, name):
        return getattr(self.__patch, name)

    def draw(self, renderer):
        oldLineStype = self.__patch.get_linestyle()
        if self.linebgcolor is not None and oldLineStype != "solid":
            oldLineColor = self.__patch.get_edgecolor()
            oldHatch = self.__patch.get_hatch()
            self.__patch.set_linestyle("solid")
            self.__patch.set_edgecolor(self.linebgcolor)
            self.__patch.set_hatch(None)
            self.__patch.draw(renderer)
            self.__patch.set_linestyle(oldLineStype)
            self.__patch.set_edgecolor(oldLineColor)
            self.__patch.set_hatch(oldHatch)
        self.__patch.draw(renderer)

    def set_transform(self, transform):
        self.__patch.set_transform(transform)

    def get_path(self):
        return self.__patch.get_path()

    def contains(self, mouseevent, radius=None):
        return self.__patch.contains(mouseevent, radius)

    def contains_point(self, point, radius=None):
        return self.__patch.contains_point(point, radius)


class Image(AxesImage):
    """An AxesImage with a fast path for uint8 RGBA images.

    :param List[float] silx_origin: (ox, oy) Offset of the image.
    :param List[float] silx_scale: (sx, sy) Scale of the image.
    """

    def __init__(self, *args,
                 silx_origin=(0., 0.),
                 silx_scale=(1., 1.),
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.__silx_origin = silx_origin
        self.__silx_scale = silx_scale

    def contains(self, mouseevent):
        """Overridden to fill 'ind' with row and column"""
        inside, info = super().contains(mouseevent)
        if inside:
            x, y = mouseevent.xdata, mouseevent.ydata
            ox, oy = self.__silx_origin
            sx, sy = self.__silx_scale
            height, width = self.get_size()
            column = numpy.clip(int((x - ox) / sx), 0, width - 1)
            row = numpy.clip(int((y - oy) / sy), 0, height - 1)
            info['ind'] = (row,), (column,)
        return inside, info

    def set_data(self, A):
        """Overridden to add a fast path for RGBA unit8 images"""
        A = numpy.array(A, copy=False)
        if A.ndim != 3 or A.shape[2] != 4 or A.dtype != numpy.uint8:
            super(Image, self).set_data(A)
        else:
            # Call AxesImage.set_data with small data to set attributes
            super(Image, self).set_data(numpy.zeros((2, 2, 4), dtype=A.dtype))
            self._A = A  # Override stored data


class BackendMatplotlib(BackendBase.BackendBase):
    """Base class for Matplotlib backend without a FigureCanvas.

    For interactive on screen plot, see :class:`BackendMatplotlibQt`.

    See :class:`BackendBase.BackendBase` for public API documentation.
    """

    def __init__(self, plot, parent=None):
        super(BackendMatplotlib, self).__init__(plot, parent)

        # matplotlib is handling keep aspect ratio at draw time
        # When keep aspect ratio is on, and one changes the limits and
        # ask them *before* next draw has been performed he will get the
        # limits without applying keep aspect ratio.
        # This attribute is used to ensure consistent values returned
        # when getting the limits at the expense of a replot
        self._dirtyLimits = True
        self._axesDisplayed = True
        self._matplotlibVersion = _parse_version(matplotlib.__version__)

        self.fig = Figure()
        self.fig.set_facecolor("w")

        self.ax = self.fig.add_axes([.15, .15, .75, .75], label="left")
        self.ax2 = self.ax.twinx()
        self.ax2.set_label("right")
        # Make sure background of Axes is displayed
        self.ax2.patch.set_visible(False)
        self.ax.patch.set_visible(True)

        # Set axis zorder=0.5 so grid is displayed at 0.5
        self.ax.set_axisbelow(True)

        # disable the use of offsets
        try:
            axes = [
                self.ax.get_yaxis().get_major_formatter(),
                self.ax.get_xaxis().get_major_formatter(),
                self.ax2.get_yaxis().get_major_formatter(),
                self.ax2.get_xaxis().get_major_formatter(),
            ]
            for axis in axes:
                axis.set_useOffset(False)
                axis.set_scientific(False)
        except:
            _logger.warning('Cannot disabled axes offsets in %s '
                            % matplotlib.__version__)

        self.ax2.set_autoscaley_on(True)

        # this works but the figure color is left
        if self._matplotlibVersion < _parse_version('2'):
            self.ax.set_axis_bgcolor('none')
        else:
            self.ax.set_facecolor('none')
        self.fig.sca(self.ax)

        self._background = None

        self._colormaps = {}

        self._graphCursor = tuple()

        self._enableAxis('right', False)
        self._isXAxisTimeSeries = False

    def getItemsFromBackToFront(self, condition=None):
        """Order as BackendBase + take into account matplotlib Axes structure"""
        def axesOrder(item):
            if item.isOverlay():
                return 2
            elif isinstance(item, items.YAxisMixIn) and item.getYAxis() == 'right':
                return 1
            else:
                return 0

        return sorted(
            BackendBase.BackendBase.getItemsFromBackToFront(
                self, condition=condition),
            key=axesOrder)

    def _overlayItems(self):
        """Generator of backend renderer for overlay items"""
        for item in self._plot.getItems():
            if (item.isOverlay() and
                    item.isVisible() and
                    item._backendRenderer is not None):
                yield item._backendRenderer

    def _hasOverlays(self):
        """Returns whether there is an overlay layer or not.

        The overlay layers contains overlay items and the crosshair.

        :rtype: bool
        """
        if self._graphCursor:
            return True  # There is the crosshair

        for item in self._overlayItems():
            return True  # There is at least one overlay item
        return False

    # Add methods

    def _getMarkerFromSymbol(self, symbol):
        """Returns a marker that can be displayed by matplotlib.

        :param str symbol: A symbol description used by silx
        :rtype: Union[str,int,matplotlib.path.Path]
        """
        path = get_path_from_symbol(symbol)
        if path is not None:
            return path
        num = _SPECIAL_MARKERS.get(symbol, None)
        if num is not None:
            return num
        # This symbol must be supported by matplotlib
        return symbol

    def addCurve(self, x, y,
                 color, symbol, linewidth, linestyle,
                 yaxis,
                 xerror, yerror,
                 fill, alpha, symbolsize, baseline):
        for parameter in (x, y, color, symbol, linewidth, linestyle,
                          yaxis, fill, alpha, symbolsize):
            assert parameter is not None
        assert yaxis in ('left', 'right')

        if (len(color) == 4 and
                type(color[3]) in [type(1), numpy.uint8, numpy.int8]):
            color = numpy.array(color, dtype=numpy.float64) / 255.

        if yaxis == "right":
            axes = self.ax2
            self._enableAxis("right", True)
        else:
            axes = self.ax

        pickradius = 3

        artists = []  # All the artists composing the curve

        # First add errorbars if any so they are behind the curve
        if xerror is not None or yerror is not None:
            if hasattr(color, 'dtype') and len(color) == len(x):
                errorbarColor = 'k'
            else:
                errorbarColor = color

            # Nx1 error array deprecated in matplotlib >=3.1 (removed in 3.3)
            if (isinstance(xerror, numpy.ndarray) and xerror.ndim == 2 and
                        xerror.shape[1] == 1):
                xerror = numpy.ravel(xerror)
            if (isinstance(yerror, numpy.ndarray) and yerror.ndim == 2 and
                    yerror.shape[1] == 1):
                yerror = numpy.ravel(yerror)

            errorbars = axes.errorbar(x, y,
                                      xerr=xerror, yerr=yerror,
                                      linestyle=' ', color=errorbarColor)
            artists += list(errorbars.get_children())

        if hasattr(color, 'dtype') and len(color) == len(x):
            # scatter plot
            if color.dtype not in [numpy.float32, numpy.float64]:
                actualColor = color / 255.
            else:
                actualColor = color

            if linestyle not in ["", " ", None]:
                # scatter plot with an actual line ...
                # we need to assign a color ...
                curveList = axes.plot(x, y,
                                      linestyle=linestyle,
                                      color=actualColor[0],
                                      linewidth=linewidth,
                                      picker=True,
                                      pickradius=pickradius,
                                      marker=None)
                artists += list(curveList)

            marker = self._getMarkerFromSymbol(symbol)
            scatter = axes.scatter(x, y,
                                   color=actualColor,
                                   marker=marker,
                                   picker=True,
                                   pickradius=pickradius,
                                   s=symbolsize**2)
            artists.append(scatter)

            if fill:
                if baseline is None:
                    _baseline = FLOAT32_MINPOS
                else:
                    _baseline = baseline
                artists.append(axes.fill_between(
                    x, _baseline, y, facecolor=actualColor[0], linestyle=''))

        else:  # Curve
            curveList = axes.plot(x, y,
                                  linestyle=linestyle,
                                  color=color,
                                  linewidth=linewidth,
                                  marker=symbol,
                                  picker=True,
                                  pickradius=pickradius,
                                  markersize=symbolsize)
            artists += list(curveList)

            if fill:
                if baseline is None:
                    _baseline = FLOAT32_MINPOS
                else:
                    _baseline = baseline
                artists.append(
                    axes.fill_between(x, _baseline, y, facecolor=color))

        for artist in artists:
            if alpha < 1:
                artist.set_alpha(alpha)

        return _PickableContainer(artists)

    def addImage(self, data, origin, scale, colormap, alpha):
        # Non-uniform image
        # http://wiki.scipy.org/Cookbook/Histograms
        # Non-linear axes
        # http://stackoverflow.com/questions/11488800/non-linear-axes-for-imshow-in-matplotlib
        for parameter in (data, origin, scale):
            assert parameter is not None

        origin = float(origin[0]), float(origin[1])
        scale = float(scale[0]), float(scale[1])
        height, width = data.shape[0:2]

        # All image are shown as RGBA image
        image = Image(self.ax,
                      interpolation='nearest',
                      picker=True,
                      origin='lower',
                      silx_origin=origin,
                      silx_scale=scale)

        if alpha < 1:
            image.set_alpha(alpha)

        # Set image extent
        xmin = origin[0]
        xmax = xmin + scale[0] * width
        if scale[0] < 0.:
            xmin, xmax = xmax, xmin

        ymin = origin[1]
        ymax = ymin + scale[1] * height
        if scale[1] < 0.:
            ymin, ymax = ymax, ymin

        image.set_extent((xmin, xmax, ymin, ymax))

        # Set image data
        if scale[0] < 0. or scale[1] < 0.:
            # For negative scale, step by -1
            xstep = 1 if scale[0] >= 0. else -1
            ystep = 1 if scale[1] >= 0. else -1
            data = data[::ystep, ::xstep]

        if data.ndim == 2:  # Data image, convert to RGBA image
            data = colormap.applyToData(data)
        elif data.dtype == numpy.uint16:
            # Normalize uint16 data to have a similar behavior as opengl backend
            data = data.astype(numpy.float32)
            data /= 65535
        
        image.set_data(data)
        self.ax.add_artist(image)
        return image

    def addTriangles(self, x, y, triangles, color, alpha):
        for parameter in (x, y, triangles, color, alpha):
            assert parameter is not None

        color = numpy.array(color, copy=False)
        assert color.ndim == 2 and len(color) == len(x)

        if color.dtype not in [numpy.float32, numpy.float64]:
            color = color.astype(numpy.float32) / 255.

        collection = TriMesh(
            Triangulation(x, y, triangles),
            alpha=alpha,
            pickradius=0)  # 0 enables picking on filled triangle
        collection.set_color(color)
        self.ax.add_collection(collection)

        return collection

    def addShape(self, x, y, shape, color, fill, overlay,
                 linestyle, linewidth, linebgcolor):
        if (linebgcolor is not None and
                shape not in ('rectangle', 'polygon', 'polylines')):
            _logger.warning(
                'linebgcolor not implemented for %s with matplotlib backend',
                shape)
        xView = numpy.array(x, copy=False)
        yView = numpy.array(y, copy=False)

        linestyle = normalize_linestyle(linestyle)

        if shape == "line":
            item = self.ax.plot(x, y, color=color,
                                linestyle=linestyle, linewidth=linewidth,
                                marker=None)[0]

        elif shape == "hline":
            if hasattr(y, "__len__"):
                y = y[-1]
            item = self.ax.axhline(y, color=color,
                                   linestyle=linestyle, linewidth=linewidth)

        elif shape == "vline":
            if hasattr(x, "__len__"):
                x = x[-1]
            item = self.ax.axvline(x, color=color,
                                   linestyle=linestyle, linewidth=linewidth)

        elif shape == 'rectangle':
            xMin = numpy.nanmin(xView)
            xMax = numpy.nanmax(xView)
            yMin = numpy.nanmin(yView)
            yMax = numpy.nanmax(yView)
            w = xMax - xMin
            h = yMax - yMin
            item = Rectangle(xy=(xMin, yMin),
                             width=w,
                             height=h,
                             fill=False,
                             color=color,
                             linestyle=linestyle,
                             linewidth=linewidth)
            if fill:
                item.set_hatch('.')

            if linestyle != "solid" and linebgcolor is not None:
                item = _DoubleColoredLinePatch(item)
                item.linebgcolor = linebgcolor

            self.ax.add_patch(item)

        elif shape in ('polygon', 'polylines'):
            points = numpy.array((xView, yView)).T
            if shape == 'polygon':
                closed = True
            else:  # shape == 'polylines'
                closed = numpy.all(numpy.equal(points[0], points[-1]))
            item = Polygon(points,
                           closed=closed,
                           fill=False,
                           color=color,
                           linestyle=linestyle,
                           linewidth=linewidth)
            if fill and shape == 'polygon':
                item.set_hatch('/')

            if linestyle != "solid" and linebgcolor is not None:
                item = _DoubleColoredLinePatch(item)
                item.linebgcolor = linebgcolor

            self.ax.add_patch(item)

        else:
            raise NotImplementedError("Unsupported item shape %s" % shape)

        if overlay:
            item.set_animated(True)

        return item

    def addMarker(self, x, y, text, color,
                  symbol, linestyle, linewidth, constraint, yaxis):
        textArtist = None

        xmin, xmax = self.getGraphXLimits()
        ymin, ymax = self.getGraphYLimits(axis=yaxis)

        if yaxis == 'left':
            ax = self.ax
        elif yaxis == 'right':
            ax = self.ax2
        else:
            assert(False)

        marker = self._getMarkerFromSymbol(symbol)
        if x is not None and y is not None:
            line = ax.plot(x, y,
                           linestyle=" ",
                           color=color,
                           marker=marker,
                           markersize=10.)[-1]

            if text is not None:
                textArtist = _TextWithOffset(x, y, text,
                                             color=color,
                                             horizontalalignment='left')
                if symbol is not None:
                    textArtist.pixel_offset = 10, 3
        elif x is not None:
            line = ax.axvline(x,
                              color=color,
                              linewidth=linewidth,
                              linestyle=linestyle)
            if text is not None:
                # Y position will be updated in updateMarkerText call
                textArtist = _TextWithOffset(x, 1., text,
                                             color=color,
                                             horizontalalignment='left',
                                             verticalalignment='top')
                textArtist.pixel_offset = 5, 3
        elif y is not None:
            line = ax.axhline(y,
                              color=color,
                              linewidth=linewidth,
                              linestyle=linestyle)

            if text is not None:
                # X position will be updated in updateMarkerText call
                textArtist = _TextWithOffset(1., y, text,
                                             color=color,
                                             horizontalalignment='right',
                                             verticalalignment='top')
                textArtist.pixel_offset = 5, 3
        else:
            raise RuntimeError('A marker must at least have one coordinate')

        line.set_picker(True)
        line.set_pickradius(5)

        # All markers are overlays
        line.set_animated(True)
        if textArtist is not None:
            ax.add_artist(textArtist)
            textArtist.set_animated(True)

        artists = [line] if textArtist is None else [line, textArtist]
        container = _MarkerContainer(artists, symbol, x, y, yaxis)
        container.updateMarkerText(xmin, xmax, ymin, ymax, self.isYAxisInverted())

        return container

    def _updateMarkers(self):
        xmin, xmax = self.ax.get_xbound()
        ymin1, ymax1 = self.ax.get_ybound()
        ymin2, ymax2 = self.ax2.get_ybound()
        yinverted = self.isYAxisInverted()
        for item in self._overlayItems():
            if isinstance(item, _MarkerContainer):
                if item.yAxis == 'left':
                    item.updateMarkerText(xmin, xmax, ymin1, ymax1, yinverted)
                else:
                    item.updateMarkerText(xmin, xmax, ymin2, ymax2, yinverted)

    # Remove methods

    def remove(self, item):
        try:
            item.remove()
        except ValueError:
            pass  # Already removed e.g., in set[X|Y]AxisLogarithmic

    # Interaction methods

    def setGraphCursor(self, flag, color, linewidth, linestyle):
        if flag:
            lineh = self.ax.axhline(
                self.ax.get_ybound()[0], visible=False, color=color,
                linewidth=linewidth, linestyle=linestyle)
            lineh.set_animated(True)

            linev = self.ax.axvline(
                self.ax.get_xbound()[0], visible=False, color=color,
                linewidth=linewidth, linestyle=linestyle)
            linev.set_animated(True)

            self._graphCursor = lineh, linev
        else:
            if self._graphCursor:
                lineh, linev = self._graphCursor
                lineh.remove()
                linev.remove()
                self._graphCursor = tuple()

    # Active curve

    def setCurveColor(self, curve, color):
        # Store Line2D and PathCollection
        for artist in curve.get_children():
            if isinstance(artist, (Line2D, LineCollection)):
                artist.set_color(color)
            elif isinstance(artist, PathCollection):
                artist.set_facecolors(color)
                artist.set_edgecolors(color)
            else:
                _logger.warning(
                    'setActiveCurve ignoring artist %s', str(artist))

    # Misc.

    def getWidgetHandle(self):
        return self.fig.canvas

    def _enableAxis(self, axis, flag=True):
        """Show/hide Y axis

        :param str axis: Axis name: 'left' or 'right'
        :param bool flag: Default, True
        """
        assert axis in ('right', 'left')
        axes = self.ax2 if axis == 'right' else self.ax
        axes.get_yaxis().set_visible(flag)

    def replot(self):
        """Do not perform rendering.

        Override in subclass to actually draw something.
        """
        # TODO images, markers? scatter plot? move in remove?
        # Right Y axis only support curve for now
        # Hide right Y axis if no line is present
        self._dirtyLimits = False
        if not self.ax2.lines:
            self._enableAxis('right', False)

    def _drawOverlays(self):
        """Draw overlays if any."""
        def condition(item):
            return (item.isVisible() and
                    item._backendRenderer is not None and
                    item.isOverlay())

        for item in self.getItemsFromBackToFront(condition=condition):
            if (isinstance(item, items.YAxisMixIn) and
                    item.getYAxis() == 'right'):
                axes = self.ax2
            else:
                axes = self.ax
            axes.draw_artist(item._backendRenderer)

        for item in self._graphCursor:
            self.ax.draw_artist(item)

    def updateZOrder(self):
        """Reorder all items with z order from 0 to 1"""
        items = self.getItemsFromBackToFront(
            lambda item: item.isVisible() and item._backendRenderer is not None)
        count = len(items)
        for index, item in enumerate(items):
            if item.getZValue() < 0.5:
                # Make sure matplotlib z order is below the grid (with z=0.5)
                zorder = 0.5 * index / count
            else:  # Make sure matplotlib z order is above the grid (> 0.5)
                zorder = 1. + index / count
            if zorder != item._backendRenderer.get_zorder():
                item._backendRenderer.set_zorder(zorder)

    def saveGraph(self, fileName, fileFormat, dpi):
        self.updateZOrder()

        # fileName can be also a StringIO or file instance
        if dpi is not None:
            self.fig.savefig(fileName, format=fileFormat, dpi=dpi)
        else:
            self.fig.savefig(fileName, format=fileFormat)
        self._plot._setDirtyPlot()

    # Graph labels

    def setGraphTitle(self, title):
        self.ax.set_title(title)

    def setGraphXLabel(self, label):
        self.ax.set_xlabel(label)

    def setGraphYLabel(self, label, axis):
        axes = self.ax if axis == 'left' else self.ax2
        axes.set_ylabel(label)

    # Graph limits

    def setLimits(self, xmin, xmax, ymin, ymax, y2min=None, y2max=None):
        # Let matplotlib taking care of keep aspect ratio if any
        self._dirtyLimits = True
        self.ax.set_xlim(min(xmin, xmax), max(xmin, xmax))

        if y2min is not None and y2max is not None:
            if not self.isYAxisInverted():
                self.ax2.set_ylim(min(y2min, y2max), max(y2min, y2max))
            else:
                self.ax2.set_ylim(max(y2min, y2max), min(y2min, y2max))

        if not self.isYAxisInverted():
            self.ax.set_ylim(min(ymin, ymax), max(ymin, ymax))
        else:
            self.ax.set_ylim(max(ymin, ymax), min(ymin, ymax))

        self._updateMarkers()

    def getGraphXLimits(self):
        if self._dirtyLimits and self.isKeepDataAspectRatio():
            self.ax.apply_aspect()
            self.ax2.apply_aspect()
            self._dirtyLimits = False
        return self.ax.get_xbound()

    def setGraphXLimits(self, xmin, xmax):
        self._dirtyLimits = True
        self.ax.set_xlim(min(xmin, xmax), max(xmin, xmax))
        self._updateMarkers()

    def getGraphYLimits(self, axis):
        assert axis in ('left', 'right')
        ax = self.ax2 if axis == 'right' else self.ax

        if not ax.get_visible():
            return None

        if self._dirtyLimits and self.isKeepDataAspectRatio():
            self.ax.apply_aspect()
            self.ax2.apply_aspect()
            self._dirtyLimits = False

        return ax.get_ybound()

    def setGraphYLimits(self, ymin, ymax, axis):
        ax = self.ax2 if axis == 'right' else self.ax
        if ymax < ymin:
            ymin, ymax = ymax, ymin
        self._dirtyLimits = True

        if self.isKeepDataAspectRatio():
            # matplotlib keeps limits of shared axis when keeping aspect ratio
            # So x limits are kept when changing y limits....
            # Change x limits first by taking into account aspect ratio
            # and then change y limits.. so matplotlib does not need
            # to make change (to y) to keep aspect ratio
            xmin, xmax = ax.get_xbound()
            curYMin, curYMax = ax.get_ybound()

            newXRange = (xmax - xmin) * (ymax - ymin) / (curYMax - curYMin)
            xcenter = 0.5 * (xmin + xmax)
            ax.set_xlim(xcenter - 0.5 * newXRange, xcenter + 0.5 * newXRange)

        if not self.isYAxisInverted():
            ax.set_ylim(ymin, ymax)
        else:
            ax.set_ylim(ymax, ymin)

        self._updateMarkers()

    # Graph axes

    def setXAxisTimeZone(self, tz):
        super(BackendMatplotlib, self).setXAxisTimeZone(tz)

        # Make new formatter and locator with the time zone.
        self.setXAxisTimeSeries(self.isXAxisTimeSeries())

    def isXAxisTimeSeries(self):
        return self._isXAxisTimeSeries

    def setXAxisTimeSeries(self, isTimeSeries):
        self._isXAxisTimeSeries = isTimeSeries
        if self._isXAxisTimeSeries:
            # We can't use a matplotlib.dates.DateFormatter because it expects
            # the data to be in datetimes. Silx works internally with
            # timestamps (floats).
            locator = NiceDateLocator(tz=self.getXAxisTimeZone())
            self.ax.xaxis.set_major_locator(locator)
            self.ax.xaxis.set_major_formatter(
                NiceAutoDateFormatter(locator, tz=self.getXAxisTimeZone()))
        else:
            try:
                scalarFormatter = ScalarFormatter(useOffset=False)
            except:
                _logger.warning('Cannot disabled axes offsets in %s ' %
                                matplotlib.__version__)
                scalarFormatter = ScalarFormatter()
            self.ax.xaxis.set_major_formatter(scalarFormatter)

    def setXAxisLogarithmic(self, flag):
        # Workaround for matplotlib 2.1.0 when one tries to set an axis
        # to log scale with both limits <= 0
        # In this case a draw with positive limits is needed first
        if flag and self._matplotlibVersion >= _parse_version('2.1.0'):
            xlim = self.ax.get_xlim()
            if xlim[0] <= 0 and xlim[1] <= 0:
                self.ax.set_xlim(1, 10)
                self.draw()

        self.ax2.set_xscale('log' if flag else 'linear')
        self.ax.set_xscale('log' if flag else 'linear')

    def setYAxisLogarithmic(self, flag):
        # Workaround for matplotlib 2.0 issue with negative bounds
        # before switching to log scale
        if flag and self._matplotlibVersion >= _parse_version('2.0.0'):
            redraw = False
            for axis, dataRangeIndex in ((self.ax, 1), (self.ax2, 2)):
                ylim = axis.get_ylim()
                if ylim[0] <= 0 or ylim[1] <= 0:
                    dataRange = self._plot.getDataRange()[dataRangeIndex]
                    if dataRange is None:
                        dataRange = 1, 100  # Fallback
                    axis.set_ylim(*dataRange)
                    redraw = True
            if redraw:
                self.draw()

        self.ax2.set_yscale('log' if flag else 'linear')
        self.ax.set_yscale('log' if flag else 'linear')

    def setYAxisInverted(self, flag):
        if self.ax.yaxis_inverted() != bool(flag):
            self.ax.invert_yaxis()
            self._updateMarkers()

    def isYAxisInverted(self):
        return self.ax.yaxis_inverted()

    def isKeepDataAspectRatio(self):
        return self.ax.get_aspect() in (1.0, 'equal')

    def setKeepDataAspectRatio(self, flag):
        self.ax.set_aspect(1.0 if flag else 'auto')
        self.ax2.set_aspect(1.0 if flag else 'auto')

    def setGraphGrid(self, which):
        self.ax.grid(False, which='both')  # Disable all grid first
        if which is not None:
            self.ax.grid(True, which=which)

    # Data <-> Pixel coordinates conversion

    def _getDevicePixelRatio(self) -> float:
        """Compatibility wrapper for devicePixelRatioF"""
        return 1.

    def _mplToQtPosition(self, x: float, y: float) -> Tuple[float, float]:
        """Convert matplotlib "display" space coord to Qt widget logical pixel
        """
        ratio = self._getDevicePixelRatio()
        # Convert from matplotlib origin (bottom) to Qt origin (top)
        # and apply device pixel ratio
        return x / ratio, (self.fig.get_window_extent().height - y) / ratio

    def _qtToMplPosition(self, x: float, y: float) -> Tuple[float, float]:
        """Convert Qt widget logical pixel to matplotlib "display" space coord
        """
        ratio = self._getDevicePixelRatio()
        # Apply device pixel ration and
        # convert from Qt origin (top) to matplotlib origin (bottom)
        return x * ratio, self.fig.get_window_extent().height - (y * ratio)

    def dataToPixel(self, x, y, axis):
        ax = self.ax2 if axis == "right" else self.ax
        displayPos = ax.transData.transform_point((x, y)).transpose()
        return self._mplToQtPosition(*displayPos)

    def pixelToData(self, x, y, axis):
        ax = self.ax2 if axis == "right" else self.ax
        displayPos = self._qtToMplPosition(x, y)
        return tuple(ax.transData.inverted().transform_point(displayPos))

    def getPlotBoundsInPixels(self):
        bbox = self.ax.get_window_extent()
        # Warning this is not returning int...
        ratio = self._getDevicePixelRatio()
        return tuple(int(value / ratio) for value in (
            bbox.xmin,
            self.fig.get_window_extent().height - bbox.ymax,
            bbox.width,
            bbox.height))

    def setAxesMargins(self, left: float, top: float, right: float, bottom: float):
        width, height = 1. - left - right, 1. - top - bottom
        position = left, bottom, width, height

        # Toggle display of axes and viewbox rect
        isFrameOn = position != (0., 0., 1., 1.)
        self.ax.set_frame_on(isFrameOn)
        self.ax2.set_frame_on(isFrameOn)

        self.ax.set_position(position)
        self.ax2.set_position(position)

        self._synchronizeBackgroundColors()
        self._synchronizeForegroundColors()
        self._plot._setDirtyPlot()

    def _synchronizeBackgroundColors(self):
        backgroundColor = self._plot.getBackgroundColor().getRgbF()

        dataBackgroundColor = self._plot.getDataBackgroundColor()
        if dataBackgroundColor.isValid():
            dataBackgroundColor = dataBackgroundColor.getRgbF()
        else:
            dataBackgroundColor = backgroundColor

        if self.ax.get_frame_on():
            self.fig.patch.set_facecolor(backgroundColor)
            if self._matplotlibVersion < _parse_version('2'):
                self.ax.set_axis_bgcolor(dataBackgroundColor)
            else:
                self.ax.set_facecolor(dataBackgroundColor)
        else:
            self.fig.patch.set_facecolor(dataBackgroundColor)

    def _synchronizeForegroundColors(self):
        foregroundColor = self._plot.getForegroundColor().getRgbF()

        gridColor = self._plot.getGridColor()
        if gridColor.isValid():
            gridColor = gridColor.getRgbF()
        else:
            gridColor = foregroundColor

        for axes in (self.ax, self.ax2):
            if axes.get_frame_on():
                axes.spines['bottom'].set_color(foregroundColor)
                axes.spines['top'].set_color(foregroundColor)
                axes.spines['right'].set_color(foregroundColor)
                axes.spines['left'].set_color(foregroundColor)
                axes.tick_params(axis='x', colors=foregroundColor)
                axes.tick_params(axis='y', colors=foregroundColor)
                axes.yaxis.label.set_color(foregroundColor)
                axes.xaxis.label.set_color(foregroundColor)
                axes.title.set_color(foregroundColor)

                for line in axes.get_xgridlines():
                    line.set_color(gridColor)

                for line in axes.get_ygridlines():
                    line.set_color(gridColor)
                # axes.grid().set_markeredgecolor(gridColor)

    def setBackgroundColors(self, backgroundColor, dataBackgroundColor):
        self._synchronizeBackgroundColors()

    def setForegroundColors(self, foregroundColor, gridColor):
        self._synchronizeForegroundColors()


class BackendMatplotlibQt(FigureCanvasQTAgg, BackendMatplotlib):
    """QWidget matplotlib backend using a QtAgg canvas.

    It adds fast overlay drawing and mouse event management.
    """

    _sigPostRedisplay = qt.Signal()
    """Signal handling automatic asynchronous replot"""

    def __init__(self, plot, parent=None):
        BackendMatplotlib.__init__(self, plot, parent)
        FigureCanvasQTAgg.__init__(self, self.fig)
        self.setParent(parent)

        self._limitsBeforeResize = None

        FigureCanvasQTAgg.setSizePolicy(
            self, qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding)
        FigureCanvasQTAgg.updateGeometry(self)

        # Make postRedisplay asynchronous using Qt signal
        self._sigPostRedisplay.connect(
            super(BackendMatplotlibQt, self).postRedisplay,
            qt.Qt.QueuedConnection)

        self._picked = None

        self.mpl_connect('button_press_event', self._onMousePress)
        self.mpl_connect('button_release_event', self._onMouseRelease)
        self.mpl_connect('motion_notify_event', self._onMouseMove)
        self.mpl_connect('scroll_event', self._onMouseWheel)

    def postRedisplay(self):
        self._sigPostRedisplay.emit()

    def _getDevicePixelRatio(self) -> float:
        """Compatibility wrapper for devicePixelRatioF"""
        if hasattr(self, 'devicePixelRatioF'):
            ratio = self.devicePixelRatioF()
        else:  # Qt < 5.6 compatibility
            ratio = float(self.devicePixelRatio())
        # Safety net: avoid returning 0
        return ratio if ratio != 0. else 1.

    # Mouse event forwarding

    _MPL_TO_PLOT_BUTTONS = {1: 'left', 2: 'middle', 3: 'right'}

    def _onMousePress(self, event):
        button = self._MPL_TO_PLOT_BUTTONS.get(event.button, None)
        if button is not None:
            x, y = self._mplToQtPosition(event.x, event.y)
            self._plot.onMousePress(int(x), int(y), button)

    def _onMouseMove(self, event):
        x, y = self._mplToQtPosition(event.x, event.y)
        if self._graphCursor:
            position = self._plot.pixelToData(
                x, y, axis='left', check=True)
            lineh, linev = self._graphCursor
            if position is not None:
                linev.set_visible(True)
                linev.set_xdata((position[0], position[0]))
                lineh.set_visible(True)
                lineh.set_ydata((position[1], position[1]))
                self._plot._setDirtyPlot(overlayOnly=True)
            elif lineh.get_visible():
                    lineh.set_visible(False)
                    linev.set_visible(False)
                    self._plot._setDirtyPlot(overlayOnly=True)
            # onMouseMove must trigger replot if dirty flag is raised

        self._plot.onMouseMove(int(x), int(y))

    def _onMouseRelease(self, event):
        button = self._MPL_TO_PLOT_BUTTONS.get(event.button, None)
        if button is not None:
            x, y = self._mplToQtPosition(event.x, event.y)
            self._plot.onMouseRelease(int(x), int(y), button)

    def _onMouseWheel(self, event):
        x, y = self._mplToQtPosition(event.x, event.y)
        self._plot.onMouseWheel(int(x), int(y), event.step)

    def leaveEvent(self, event):
        """QWidget event handler"""
        try:
            plot = self._plot
        except RuntimeError:
            pass
        else:
            plot.onMouseLeaveWidget()

    # picking

    def pickItem(self, x, y, item):
        xDisplay, yDisplay = self._qtToMplPosition(x, y)
        mouseEvent = MouseEvent(
            'button_press_event', self, int(xDisplay), int(yDisplay))
        # Override axes and data position with the axes
        mouseEvent.inaxes = item.axes
        mouseEvent.xdata, mouseEvent.ydata = self.pixelToData(
            x, y, axis='left' if item.axes is self.ax else 'right')
        picked, info = item.contains(mouseEvent)

        if not picked:
            return None

        elif isinstance(item, TriMesh):
            # Convert selected triangle to data point indices
            triangulation = item._triangulation
            indices = triangulation.get_masked_triangles()[info['ind'][0]]

            # Sort picked triangle points by distance to mouse
            # from furthest to closest to put closest point last
            # This is to be somewhat consistent with last scatter point
            # being the top one.
            xdata, ydata = self.pixelToData(x, y, axis='left')
            dists = ((triangulation.x[indices] - xdata) ** 2 +
                     (triangulation.y[indices] - ydata) ** 2)
            return indices[numpy.flip(numpy.argsort(dists), axis=0)]

        else:  # Returns indices if any
            return info.get('ind', ())

    # replot control

    def resizeEvent(self, event):
        # Store current limits
        self._limitsBeforeResize = (
            self.ax.get_xbound(), self.ax.get_ybound(), self.ax2.get_ybound())

        FigureCanvasQTAgg.resizeEvent(self, event)
        if self.isKeepDataAspectRatio() or self._hasOverlays():
            # This is needed with matplotlib 1.5.x and 2.0.x
            self._plot._setDirtyPlot()

    def draw(self):
        """Overload draw

        It performs a full redraw (including overlays) of the plot.
        It also resets background and emit limits changed signal.

        This is directly called by matplotlib for widget resize.
        """
        self.updateZOrder()

        # Starting with mpl 2.1.0, toggling autoscale raises a ValueError
        # in some situations. See #1081, #1136, #1163,
        if self._matplotlibVersion >= _parse_version("2.0.0"):
            try:
                FigureCanvasQTAgg.draw(self)
            except ValueError as err:
                _logger.debug(
                    "ValueError caught while calling FigureCanvasQTAgg.draw: "
                    "'%s'", err)
        else:
            FigureCanvasQTAgg.draw(self)

        if self._hasOverlays():
            # Save background
            self._background = self.copy_from_bbox(self.fig.bbox)
        else:
            self._background = None  # Reset background

        # Check if limits changed due to a resize of the widget
        if self._limitsBeforeResize is not None:
            xLimits, yLimits, yRightLimits = self._limitsBeforeResize
            self._limitsBeforeResize = None

            if (xLimits != self.ax.get_xbound() or
                    yLimits != self.ax.get_ybound()):
                self._updateMarkers()

            if xLimits != self.ax.get_xbound():
                self._plot.getXAxis()._emitLimitsChanged()
            if yLimits != self.ax.get_ybound():
                self._plot.getYAxis(axis='left')._emitLimitsChanged()
            if yRightLimits != self.ax2.get_ybound():
                self._plot.getYAxis(axis='right')._emitLimitsChanged()

        self._drawOverlays()

    def replot(self):
        BackendMatplotlib.replot(self)

        dirtyFlag = self._plot._getDirtyPlot()

        if dirtyFlag == 'overlay':
            # Only redraw overlays using fast rendering path
            if self._background is None:
                self._background = self.copy_from_bbox(self.fig.bbox)
            self.restore_region(self._background)
            self._drawOverlays()
            self.blit(self.fig.bbox)

        elif dirtyFlag:  # Need full redraw
            self.draw()

        # Workaround issue of rendering overlays with some matplotlib versions
        if (_parse_version('1.5') <= self._matplotlibVersion < _parse_version('2.1') and
                not hasattr(self, '_firstReplot')):
            self._firstReplot = False
            if self._hasOverlays():
                qt.QTimer.singleShot(0, self.draw)  # Request async draw

    # cursor

    _QT_CURSORS = {
        BackendBase.CURSOR_DEFAULT: qt.Qt.ArrowCursor,
        BackendBase.CURSOR_POINTING: qt.Qt.PointingHandCursor,
        BackendBase.CURSOR_SIZE_HOR: qt.Qt.SizeHorCursor,
        BackendBase.CURSOR_SIZE_VER: qt.Qt.SizeVerCursor,
        BackendBase.CURSOR_SIZE_ALL: qt.Qt.SizeAllCursor,
    }

    def setGraphCursorShape(self, cursor):
        if cursor is None:
            FigureCanvasQTAgg.unsetCursor(self)
        else:
            cursor = self._QT_CURSORS[cursor]
            FigureCanvasQTAgg.setCursor(self, qt.QCursor(cursor))



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Module testing silx.app.convert"""

__authors__ = ["P. Knobel"]
__license__ = "MIT"
__date__ = "17/01/2018"


import os
import sys
import tempfile
import unittest
import io
import gc
import h5py

import silx
from .. import convert
from silx.utils import testutils
from silx.io.utils import h5py_read_dataset


# content of a spec file
sftext = """#F /tmp/sf.dat
#E 1455180875
#D Thu Feb 11 09:54:35 2016
#C imaging  User = opid17
#O0 Pslit HGap  MRTSlit UP  MRTSlit DOWN
#O1 Sslit1 VOff  Sslit1 HOff  Sslit1 VGap
#o0 pshg mrtu mrtd
#o2 ss1vo ss1ho ss1vg

#J0 Seconds  IA  ion.mono  Current
#J1 xbpmc2  idgap1  Inorm

#S 1  ascan  ss1vo -4.55687 -0.556875  40 0.2
#D Thu Feb 11 09:55:20 2016
#T 0.2  (Seconds)
#P0 180.005 -0.66875 0.87125
#P1 14.74255 16.197579 12.238283
#N 4
#L MRTSlit UP  second column  3rd_col
-1.23 5.89  8
8.478100E+01  5 1.56
3.14 2.73 -3.14
1.2 2.3 3.4

#S 1 aaaaaa
#D Thu Feb 11 10:00:32 2016
#@MCADEV 1
#@MCA %16C
#@CHANN 3 0 2 1
#@CALIB 1 2 3
#N 3
#L uno  duo
1 2
@A 0 1 2
@A 10 9 8
3 4
@A 3.1 4 5
@A 7 6 5
5 6
@A 6 7.7 8
@A 4 3 2
"""


class TestConvertCommand(unittest.TestCase):
    """Test command line parsing"""

    def testHelp(self):
        # option -h must cause a `raise SystemExit` or a `return 0`
        try:
            result = convert.main(["convert", "--help"])
        except SystemExit as e:
            result = e.args[0]
        self.assertEqual(result, 0)

    def testWrongOption(self):
        # presence of a wrong option must cause a SystemExit or a return
        # with a non-zero status
        try:
            result = convert.main(["convert", "--foo"])
        except SystemExit as e:
            result = e.args[0]
        self.assertNotEqual(result, 0)

    @testutils.test_logging(convert._logger.name, error=3)
    # one error log per missing file + one "Aborted" error log
    def testWrongFiles(self):
        result = convert.main(["convert", "foo.spec", "bar.edf"])
        self.assertNotEqual(result, 0)

    def testFile(self):
        # create a writable temp directory
        tempdir = tempfile.mkdtemp()

        # write a temporary SPEC file
        specname = os.path.join(tempdir, "input.dat")
        with io.open(specname, "wb") as fd:
            if sys.version_info < (3, ):
                fd.write(sftext)
            else:
                fd.write(bytes(sftext, 'ascii'))

        # convert it
        h5name = os.path.join(tempdir, "output.h5")
        assert not os.path.isfile(h5name)
        command_list = ["convert", "-m", "w",
                        specname, "-o", h5name]
        result = convert.main(command_list)

        self.assertEqual(result, 0)
        self.assertTrue(os.path.isfile(h5name))

        with h5py.File(h5name, "r") as h5f:
            title12 = h5py_read_dataset(h5f["/1.2/title"])
            if sys.version_info < (3, ):
                title12 = title12.encode("utf-8")
            self.assertEqual(title12,
                             "aaaaaa")

            creator = h5f.attrs.get("creator")
            self.assertIsNotNone(creator, "No creator attribute in NXroot group")
            if sys.version_info < (3, ):
                creator = creator.encode("utf-8")
            self.assertIn("silx convert (v%s)" % silx.version, creator)

        # delete input file
        gc.collect()  # necessary to free spec file on Windows
        os.unlink(specname)
        os.unlink(h5name)
        os.rmdir(tempdir)


def suite():
    test_suite = unittest.TestSuite()
    loader = unittest.defaultTestLoader.loadTestsFromTestCase
    test_suite.addTest(loader(TestConvertCommand))
    return test_suite


if __name__ == '__main__':
    unittest.main(defaultTest='suite')
